use std::ffi::OsStr;
use std::fs::File;
use std::path::PathBuf;

use anyhow::{Error, Result};
use chrono::{DateTime, Utc};
use env_logger::{Builder, Env};
use semver::Version;
// use serde::export::TryFrom;
use std::convert::TryFrom;
use serde::Deserialize;
use structopt::StructOpt;

use rusty_hektor::input_types::ilias::from_zip;
use rusty_hektor::input_types::ilias::results_csv::results_mapping_from_path;
use rusty_hektor::input_types::various::email_group_data_from_path;
use rusty_hektor::io::yes_or_no;
use rusty_hektor::post_process::compile_test::CompileTest;
use rusty_hektor::post_process::empty_test::EmptyTest;
use rusty_hektor::post_process::Test;
use rusty_hektor::post_process::{post_process, PostProcessContext};
use rusty_hektor::transform::{merge_transform, transform, TransformContext};

// TODO this should really not be here but in the main
/// Parse ILIAS exam export for importing into Grady
#[derive(Debug, StructOpt)]
pub struct Opt {
    /// Provide the path to the zipped xml, if the exam was split into multiple
    /// cohorts with the same module and exercises, you can provide multiple .zip exports.
    /// If provided exercises are not identical, a warning is issued and the program falls back
    /// to the exercises of the first export file
    #[structopt(name = "IN_FILE", required = true, parse(from_os_str))]
    in_files: Vec<PathBuf>,

    /// When processing a single input file, a results csv file (generated by ILIAS when
    /// selecting "Erstelle Exportdatei") can be processed in order to map students to their
    /// real names and usernames. Currently only needed when processing exercise sessions.
    #[structopt(long, name = "RESULTS_CSV", parse(from_os_str))]
    results_csv: Option<PathBuf>,

    /// Where to store the parsed data, prints to stdout if not provided
    #[structopt(short, long, name = "OUT_FILE", parse(from_os_str))]
    out_file: Option<PathBuf>,

    /// Use anonymisation
    #[structopt(short, long)]
    anon: bool,

    /// Explicitly set passwords for users
    #[structopt(short = "p", long)]
    set_passwords: bool,

    /// Parse cloze type questions
    #[structopt(long = "parse-cloze")]
    parse_cloze: bool,

    /// Parse cloze type questions
    #[structopt(long, default_value = "0")]
    cloze_question_field: u32,

    /// Disable latex rendering
    #[structopt(long)]
    no_latex: bool,

    /// Disable user input prompts, will use default values for all unknowns.
    /// Only intended for testing purposes.
    #[structopt(long)]
    no_user_input: bool,

    /// UNSTABLE: Add email and group information to participants by supplying a suitable mapping
    /// NOTE: This is momentarily used to utilize the group data provided by the lecturer
    /// for the correction of the cs1 exercises. It is not clear if the format of the csv
    /// file is stable or not
    #[structopt(long, name = "CSV-FILE", parse(from_os_str))]
    email_group_csv: Option<PathBuf>,

    /// Tests to run on the the submissions, the default
    /// is "empty" to check for empty submissions. Additionally a compile test
    /// for c programs can be run by also passing `compile_c`
    #[structopt(
        short,
        long,
        name = "TEST",
        parse(from_os_str),
        default_value = "empty"
    )]
    tests: Vec<TestEnum>,

    /// Directory containing header files when running c compile tests
    #[structopt(long, name = "HEADER-DIR", parse(from_os_str))]
    header_dir: Option<PathBuf>,

    /// Where to store the anonymisation map file, if enabled
    #[structopt(
        short,
        long = "map-file",
        name = "MAP_FILE_PATH",
        default_value = "anon-mapping.json"
    )]
    map_file_name: PathBuf,
}

impl std::convert::TryFrom<&Opt> for TransformContext {
    type Error = Error;

    fn try_from(opt: &Opt) -> Result<Self, Self::Error> {
        let results_csv = opt
            .results_csv
            .as_ref()
            .map(results_mapping_from_path)
            .transpose()?;
        let email_group_mapping = opt
            .email_group_csv
            .as_ref()
            .map(email_group_data_from_path)
            .transpose()?;

        let parse_cloze_questions = if opt.parse_cloze {
            Some(opt.cloze_question_field)
        } else {
            None
        };

        Ok(Self {
            results_csv,
            email_group_mapping,
            parse_cloze_questions,
            enable_latex_rendering: !opt.no_latex,
            sub_types_by_name: None,
            no_user_input: opt.no_user_input
        })
    }
}

impl From<&Opt> for PostProcessContext {
    fn from(opt: &Opt) -> Self {
        let tests = opt
            .tests
            .iter()
            .map(|test_enum| test_enum.as_test(opt))
            .collect();
        Self {
            anonymize: opt.anon,
            tests,
        }
    }
}

#[derive(Clone, Debug)]
enum TestEnum {
    Empty,
    CompileC,
}

impl TestEnum {
    fn as_test(&self, opt: &Opt) -> Box<dyn Test> {
        match self {
            Self::Empty => Box::new(EmptyTest {}),
            Self::CompileC => {
                Box::new(CompileTest::new(opt.header_dir.clone().expect(
                    "--header-dir needs to be provided when executing c compile test",
                )))
            }
        }
    }
}

impl From<&OsStr> for TestEnum {
    fn from(s: &OsStr) -> Self {
        match s.to_str().expect("Non UTF-8 string in tests") {
            "empty" => Self::Empty,
            "compile_c" => Self::CompileC,
            _ => panic!("Unable to parse test. Allowed values: empty, compile_c"),
        }
    }
}

fn main() -> Result<()> {
    setup_logging();
    let opt: Opt = Opt::from_args();
    if let Err(err) = check_for_new_version() {
        eprintln!(
            "Unable to check for a new Rusty-Hektor version.\n\
            Please check manually at {} .\n\
            Cause:\n{}",
            env!("CARGO_PKG_REPOSITORY"),
            err
        );
        if yes_or_no("Exit?") {
            std::process::exit(0)
        }
    }

    let mut transform_context = TransformContext::try_from(&opt)?;
    let post_process_context = PostProcessContext::from(&opt);

    let mut ilias_export_data: Vec<_> = opt
        .in_files
        .into_iter()
        .map(from_zip)
        .collect::<Result<Vec<_>>>()?;

    let transformed = match ilias_export_data.len() {
        1 => transform(ilias_export_data.pop().unwrap(), &transform_context)?,
        2..=std::usize::MAX => merge_transform(ilias_export_data, &mut transform_context)?,
        _ => unreachable!("Input files must be provided"),
    };

    let (post_processed, anon_mapping) = post_process(transformed, &post_process_context)?;

    if let Some(out_path) = opt.out_file {
        let out_file = File::create(out_path)?;
        serde_json::to_writer_pretty(out_file, &post_processed)?;
    } else {
        serde_json::to_writer_pretty(std::io::stdout(), &post_processed)?;
    }

    if let Some(anon_mapping) = anon_mapping {
        let out_file = File::create(opt.map_file_name)?;
        serde_json::to_writer_pretty(out_file, &anon_mapping)?;
    }

    Ok(())
}

fn setup_logging() {
    Builder::from_env(Env::default().default_filter_or("warn")).init();
}

#[derive(Deserialize, Debug)]
struct Release {
    tag_name: Version,
    created_at: DateTime<Utc>,
}

fn check_for_new_version() -> Result<()> {
    let request_url = format!(
        "https://gitlab.gwdg.de/api/v4/projects/{id}/releases",
        id = "grady-corp%2Frusty-hektor"
    );
    let response = reqwest::blocking::get(&request_url)?;
    let mut releases: Vec<Release> = response.json()?;
    releases.sort_by(|a, b| b.tag_name.cmp(&a.tag_name));
    let latest_version = &releases[0].tag_name;
    let current_version = Version::parse(env!("CARGO_PKG_VERSION"))
        .expect("This version has an invalid semver version number");
    if latest_version > &current_version {
        println!(
            "There is a new version of Rusty-Hektor available.\n\
             Your current version is {}, latest is {}.\n\
             Please download the latest version from https://gitlab.gwdg.de/robinwilliam.hundt/rusty-hektor",
            current_version, latest_version
        );
        if yes_or_no("Exit?") {
            std::process::exit(0);
        }
    }
    Ok(())
}
