use std::fs::File;
use std::io::BufReader;
use std::path::Path;

use quick_xml::de::from_reader;
use serde::de::DeserializeOwned;

use anyhow::Result;
pub use ilias::IliasExportData;

pub mod ilias;
pub mod ipynb;

pub fn from_path<P: AsRef<Path>, T: DeserializeOwned>(path: P) -> Result<T> {
    let reader = BufReader::new(File::open(path).unwrap());
    let data: T = from_reader(reader)?;
    Ok(data)
}

pub mod various {
    use std::collections::HashMap;
    use std::path::Path;

    use anyhow::Result;
    use serde::Deserialize;

    // this is used to deserialize the csv file provided for the cs1 exercises by @hbrosen
    // it is unclear if this was a one of file generated by him or if it is an export of
    // Flexnow or some other system
    /// Contains information about emails and group data
    #[derive(Deserialize, Debug)]
    pub struct EmailGroupDataRow {
        #[serde(rename = "Ilias")]
        pub ilias_username: String,
        #[serde(rename = "Studip")]
        pub studip_username: String,
        #[serde(rename = "Vorname")]
        pub first_name: String,
        #[serde(rename = "Name")]
        pub last_name: String,
        #[serde(rename = "Email")]
        pub email: String,
        #[serde(rename = "Uebung")]
        pub exercise_group: String,
    }

    /// Maps fullnames, meaning "<last_name>, <first_name>" to email and group data
    pub type EmailGroupMapping = HashMap<String, EmailGroupDataRow>;

    pub fn email_group_data_from_path(path: impl AsRef<Path>) -> Result<EmailGroupMapping> {
        let mut csv_reader = csv::ReaderBuilder::new().delimiter(b';').from_path(path)?;
        let mut mapping = HashMap::new();
        for record in csv_reader.deserialize() {
            let record: EmailGroupDataRow = record?;
            let last_name = &record.last_name;
            let first_name = &record.first_name;
            let fullname = format!("{}, {}", last_name, first_name);
            match mapping.get(&fullname) {
                None => {
                    mapping.insert(fullname, record);
                }
                Some(_) => {
                    log::error!(
                        "Student \"{}\" appears twice in email/group mapping data. Email and Group \
                        will be the same for persons with this name. FIX MANUALLY IN OUTPUT FILE!",
                        &fullname
                    );
                }
            }
        }
        Ok(mapping)
    }

    #[cfg(test)]
    mod tests {
        use super::*;

        #[test]
        fn can_parse_results_csv() {
            let _data = email_group_data_from_path(
                "src/input_types/test_data/grady-uebungen-teilnehmer.csv",
            )
            .unwrap();
        }
    }
}
